import 'dart:async';

class Store {
  Map<Type, dynamic> _current;
  final StreamController<Store> _changeController;

  /// Creates a new Store from the list of [stores]
  Store(
    List<dynamic> stores, {
    bool syncStream = true,
  }) : _changeController = StreamController.broadcast(sync: syncStream) {
    _current =
        Map.unmodifiable({for (var store in stores) store.runtimeType: store});
  }

  /// Gets the current unmodifiable state
  Map<Type, dynamic> get current => _current;

  /// Gets the Type's [T] current state
  T getState<T>() => _current[T];

  /// Sets the Type's [T] value without adding the new state to the [StreamController]
  Store unsafeSetState<T>(T value) {
    final newState = {..._current};
    newState[T] = value;
    _current = Map.unmodifiable(newState);
    return this;
  }

  /// Sets the Type's [T] value add adds the new state to the [StreamController]
  Store setState<T>(T value) => unsafeSetState<T>(value).forceEmit();

  /// Calls the updateFn on the current Type's [T] state and uses the return value as the next state
  /// adding the result to the [StreamController]
  Store updateState<T>(T Function(T) updateFn) =>
      setState<T>(updateFn(getState<T>()));

  /// Calls the updateFn on the current Type's [T] state and uses the return value as the next state
  /// without adding the result to the [StreamController]
  Store unsafeUpdateState<T>(T Function(T) updateFn) =>
      unsafeSetState<T>(updateFn(getState<T>()));

  /// Returns stream on the [StreamController]
  Stream<Store> get onChange => _changeController.stream;

  /// Converts the current state to Map<String, dynamic>
  /// each state should implement `Map<String, dynamic> toJson()` to use this method
  Map<String, dynamic> toJson() =>
      _current.map((key, value) => MapEntry(key.toString(), value.toJson()));

  /// Clears the current state and closes the [StreamController]
  Future teardown() {
    _current = Map.identity();
    return _changeController.close();
  }

  /// Forces the [StreamController] to add the current state to the stream
  Store forceEmit() {
    _changeController.add(this);
    return this;
  }
}
